import { StateField } from '@codemirror/next/state';
import { linter } from '@codemirror/next/lint';
export const slowUpdateLinter = (connection, {
  slowUpdateResult
}) => {
  let resolveDiagnostics;

  const mapDiagnostic = ({
    span: {
      start,
      length
    },
    message,
    severity
  }) => ({
    from: start,
    to: start + length,
    message,
    severity: severity
  });

  connection.on('message', message => {
    if (message.type !== 'slowUpdate') return;

    if (resolveDiagnostics) {
      const diagnostics = message.diagnostics.map(mapDiagnostic);
      diagnostics.sort((a, b) => {
        if (a.from > b.from) return 1;
        if (b.from > a.from) return -1;
        return 0;
      });
      resolveDiagnostics(diagnostics);
    }

    if (slowUpdateResult) slowUpdateResult(message);
  });
  const active = StateField.define({
    create(state) {
      return state.doc.length > 0;
    },

    update(value, _, newState) {
      return value || newState.doc.length > 0;
    }

  });
  return [active, linter(async view => {
    if (!view.state.field(active)) return [];
    const promise = new Promise(r => resolveDiagnostics = r);
    await connection.sendSlowUpdate();
    return promise;
  })];
};