const _Editor_lintingSuspended = Symbol("#lintingSuspended"),
      _Editor_hadChangesSinceLastLinting = Symbol("#hadChangesSinceLastLinting"),
      _Editor_changePending = Symbol("#changePending"),
      _Editor_changeReason = Symbol("#changeReason"),
      _Editor_changesAreFromServer = Symbol("#changesAreFromServer"),
      _Editor_onConnectionOpen = Symbol("#onConnectionOpen"),
      _Editor_hideConnectionLoss = Symbol("#hideConnectionLoss"),
      _Editor_shouldResendServerOptions = Symbol("#shouldResendServerOptions"),
      _Editor_hasDefaultServerOptions = Symbol("#hasDefaultServerOptions"),
      _Editor_connection = Symbol("#connection"),
      _Editor_serverOptions = Symbol("#serverOptions"),
      _Editor_getCursorIndex = Symbol("#getCursorIndex"),
      _Editor_options = Symbol("#options"),
      _Editor_onConnectionMessage = Symbol("#onConnectionMessage"),
      _Editor_receiveServerChanges = Symbol("#receiveServerChanges"),
      _Editor_showSlowUpdate = Symbol("#showSlowUpdate"),
      _Editor_receiveServerOptions = Symbol("#receiveServerOptions"),
      _Editor_onConnectionCloseOrError = Symbol("#onConnectionCloseOrError"),
      _Editor_showConnectionLoss = Symbol("#showConnectionLoss"),
      _Editor_cmView = Symbol("#cmView"),
      _Editor_requestSlowUpdate = Symbol("#requestSlowUpdate"),
      _Editor_connectionLossElement = Symbol("#connectionLossElement"),
      _Editor_wrapper = Symbol("#wrapper"),
      _Editor_sendServerOptions = Symbol("#sendServerOptions"),
      _Editor_language = Symbol("#language"),
      _Editor_hinter = Symbol("#hinter"),
      _Editor_signatureTip = Symbol("#signatureTip"),
      _Editor_removeConnectionEvents = Symbol("#removeConnectionEvents");

import { EditorView } from '@codemirror/next/view'; // import type { SelfDebug } from './self-debug';

import { createState } from "./codemirror/create-state.js"; // import { renderInfotip } from './render-infotip';

import { Hinter } from "./hinter.js";
import { SignatureTip } from "./signature-tip.js";
import { addEvents } from "../helpers/add-events.js"; // const languageModes = {
//     'C#': 'text/x-csharp',
//     'Visual Basic': 'text/x-vb',
//     'F#': 'text/x-fsharp',
//     'PHP': 'application/x-httpd-php'
// } as const;

const defaultLanguage = 'C#'; // const lineSeparator = '\r\n';

export class Editor {
  constructor(container, connection, // selfDebug: SelfDebug|null,
  options) {
    this[_Editor_lintingSuspended] = true;
    this[_Editor_hadChangesSinceLastLinting] = false; // #capturedUpdateLinting: CodeMirror.UpdateLintingCallback|null|undefined;

    this[_Editor_changePending] = false;
    this[_Editor_changeReason] = null;
    this[_Editor_changesAreFromServer] = false;

    this[_Editor_onConnectionOpen] = e => {
      this[_Editor_hideConnectionLoss]();

      if (this[_Editor_shouldResendServerOptions]) {
        if (!this[_Editor_hasDefaultServerOptions]()) {
          // eslint-disable-next-line @typescript-eslint/no-floating-promises
          this[_Editor_connection].sendSetOptions(this[_Editor_serverOptions]);
        }
      } else {
        this[_Editor_shouldResendServerOptions] = true; // next open
      }

      const text = this.getText();

      if (text === '' || text == null) {
        this[_Editor_lintingSuspended] = false;
        return;
      } // eslint-disable-next-line @typescript-eslint/no-floating-promises


      this[_Editor_connection].sendReplaceText(0, 0, text, this[_Editor_getCursorIndex]()); // eslint-disable-next-line @typescript-eslint/no-non-null-assertion


      this[_Editor_options].on.connectionChange('open', e);

      this[_Editor_lintingSuspended] = false;
      this[_Editor_hadChangesSinceLastLinting] = true; // if (this.#capturedUpdateLinting) {
      //     // eslint-disable-next-line @typescript-eslint/no-floating-promises
      //     this.#requestSlowUpdate();
      // }
    };

    this[_Editor_onConnectionMessage] = message => {
      switch (message.type) {
        case 'changes':
          this[_Editor_receiveServerChanges](message.changes, message.reason);

          break;

        case 'completions':
          // this.#hinter.start(message.completions, message.span, {
          //     commitChars: message.commitChars,
          //     suggestion: message.suggestion
          // });
          break;

        case 'completionInfo':
          // this.#hinter.showTip(message.index, message.parts);
          break;

        case 'signatures':
          // this.#signatureTip.update(message);
          break;

        case 'infotip':
          if (!message.sections) {
            // this.#cm.infotipUpdate(null);
            return;
          } // this.#cm.infotipUpdate({
          //     data: message,
          //     range: this.#spanToRange(message.span)
          // });


          break;

        case 'slowUpdate':
          this[_Editor_showSlowUpdate](message);

          break;

        case 'optionsEcho':
          this[_Editor_receiveServerOptions](message.options);

          break;

        case 'self:debug':
          // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
          // this.#selfDebug!.displayData(message);
          break;

        case 'error':
          // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
          this[_Editor_options].on.serverError(message.message);

          break;

        default:
          throw new Error('Unknown message type "' + message.type);
      }
    };

    this[_Editor_onConnectionCloseOrError] = e => {
      this[_Editor_lintingSuspended] = true;

      this[_Editor_showConnectionLoss](); // eslint-disable-next-line @typescript-eslint/no-non-null-assertion


      const connectionChange = this[_Editor_options].on.connectionChange;

      if (e instanceof CloseEvent) {
        connectionChange('close', e);
      } else {
        connectionChange('error', e);
      }
    }; // #onCodeMirrorBeforeChange = (_: CodeMirror.Editor, change: CodeMirror.EditorChange) => {
    //     (change.from)[indexKey] = this.#cm.indexFromPos(change.from);
    //     (change.to)[indexKey] = this.#cm.indexFromPos(change.to);
    //     this.#changePending = true;
    // };
    // #onCodeMirrorCursorActivity = () => {
    //     if (this.#changePending)
    //         return;
    //     // eslint-disable-next-line @typescript-eslint/no-floating-promises
    //     this.#connection.sendMoveCursor(this.#getCursorIndex());
    // };
    // #onCodeMirrorChanges = (_: CodeMirror.Editor, changes: ReadonlyArray<CodeMirror.EditorChange>) => {
    //     this.#hadChangesSinceLastLinting = true;
    //     this.#changePending = false;
    //     const cursorIndex = this.#getCursorIndex();
    //     changes = this.#mergeChanges(changes);
    //     for (let i = 0; i < changes.length; i++) {
    //         const change = changes[i];
    //         const start = (change.from)[indexKey];
    //         const length = (change.to)[indexKey] - start;
    //         const text = change.text.join(lineSeparator);
    //         if (cursorIndex === start + 1 && text.length === 1 && !this.#changesAreFromServer) {
    //             if (length > 0) {
    //                 // eslint-disable-next-line @typescript-eslint/no-floating-promises
    //                 this.#connection.sendReplaceText(start, length, '', cursorIndex - 1);
    //             }
    //             // eslint-disable-next-line @typescript-eslint/no-floating-promises
    //             this.#connection.sendTypeChar(text);
    //         }
    //         else {
    //             // eslint-disable-next-line @typescript-eslint/no-floating-promises
    //             this.#connection.sendReplaceText(start, length, text, cursorIndex, this.#changeReason);
    //         }
    //     }
    //     // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
    //     this.#options.on!.textChange!(this.#getText);
    // };
    // #mergeChanges = (changes: ReadonlyArray<CodeMirror.EditorChange>) => {
    //     if (changes.length < 2)
    //         return changes;
    //     const canBeMerged = (first: CodeMirror.EditorChange|null, second: CodeMirror.EditorChange|null) => {
    //         return first && second
    //             && first.origin === 'undo'
    //             && second.origin === 'undo'
    //             && first.to.line === second.from.line
    //             && first.text.length === 1
    //             && second.text.length === 1
    //             && second.from.ch === second.to.ch
    //             && (first.to.ch + first.text[0].length) === second.from.ch;
    //     };
    //     const results = [];
    //     let before: CodeMirror.EditorChange|null = null;
    //     for (const change of changes) {
    //         if (canBeMerged(before, change)) {
    //             before = {
    //                 /* eslint-disable @typescript-eslint/no-non-null-assertion */
    //                 from: before.from,
    //                 to: before.to,
    //                 text: [before.text[0] + change.text[0]],
    //                 origin: change.origin
    //                 /* eslint-enable @typescript-eslint/no-non-null-assertion */
    //             };
    //         }
    //         else {
    //             if (before)
    //                 results.push(before);
    //             before = change;
    //         }
    //     }
    //     // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
    //     results.push(before);
    //     return results;
    // };
    // #lintGetAnnotations = (_: string, updateLinting: CodeMirror.UpdateLintingCallback) => {
    //     if (!this.#capturedUpdateLinting) {
    //         this.#capturedUpdateLinting = function(this: ThisParameterType<CodeMirror.UpdateLintingCallback>, ...args: Parameters<CodeMirror.UpdateLintingCallback>) {
    //             const [cm] = args;
    //             // see https://github.com/codemirror/CodeMirror/blob/dbaf6a94f1ae50d387fa77893cf6b886988c2147/addon/lint/lint.js#L133
    //             // ensures that next 'id' will always match 'waitingFor'
    //             (cm.state as { lint: { waitingFor: number } }).lint.waitingFor = -1;
    //             updateLinting.apply(this, args);
    //         };
    //     }
    //     // eslint-disable-next-line @typescript-eslint/no-floating-promises
    //     this.#requestSlowUpdate();
    // };


    this[_Editor_getCursorIndex] = () => this[_Editor_cmView].state.selection.primary.from;

    this[_Editor_receiveServerChanges] = (changes, reason) => {
      this[_Editor_changesAreFromServer] = true;
      this[_Editor_changeReason] = reason ?? 'server';
      /*this.#cm.operation(() => {
          let offset = 0;
          for (const change of changes) {
              const from = this.#cm.posFromIndex(change.start + offset);
              const to = change.length > 0 ? this.#cm.posFromIndex(change.start + offset + change.length) : from;
              this.#cm.replaceRange(change.text, from, to, '+server');
              offset += change.text.length - change.length;
          }
      });*/

      this[_Editor_changeReason] = null;
      this[_Editor_changesAreFromServer] = false;
    }; // #getLintFixes = (cm: CodeMirror.Editor, line: number, annotations: ReadonlyArray<CodeMirror.Annotation>) => {
    //     const requestApplyFix = (cm: CodeMirror.Editor, line: number, fix: AnnotationFixWithId) => {
    //         // eslint-disable-next-line @typescript-eslint/no-floating-promises
    //         this.#connection.sendApplyDiagnosticAction(fix.id);
    //     };
    //     const fixes: Array<CodeMirror.AnnotationFix> = [];
    //     for (const annotation of annotations) {
    //         const diagnostic = (annotation).diagnostic;
    //         if (!diagnostic.actions)
    //             continue;
    //         for (const action of diagnostic.actions) {
    //             fixes.push({
    //                 text: action.title,
    //                 apply: requestApplyFix,
    //                 id: action.id
    //             } as AnnotationFixWithId);
    //         }
    //     }
    //     return fixes;
    // };
    // #infotipGetInfo = (cm: CodeMirror.Editor, position: CodeMirror.Position) => {
    //     // eslint-disable-next-line @typescript-eslint/no-floating-promises
    //     this.#connection.sendRequestInfoTip(cm.indexFromPos(position));
    // };


    this[_Editor_requestSlowUpdate] = force => {
      if (this[_Editor_lintingSuspended] || !(this[_Editor_hadChangesSinceLastLinting] || force)) return null;
      this[_Editor_hadChangesSinceLastLinting] = false; // eslint-disable-next-line @typescript-eslint/no-non-null-assertion

      this[_Editor_options].on.slowUpdateWait();

      return this[_Editor_connection].sendSlowUpdate();
    };

    this[_Editor_showSlowUpdate] = update => {
      // const annotations: Array<DiagnosticAnnotation> = [];
      // Higher severities must go last -- CodeMirror uses last one for the icon.
      // Unless one is error, in which case it's always error -- but still makes
      // sense to handle this consistently.
      const priorityBySeverity = {
        hidden: 0,
        info: 1,
        warning: 2,
        error: 3
      };
      const diagnostics = update.diagnostics.slice(0);
      diagnostics.sort((a, b) => {
        const aOrder = priorityBySeverity[a.severity];
        const bOrder = priorityBySeverity[b.severity];
        return aOrder !== bOrder ? aOrder > bOrder ? 1 : -1 : 0;
      });

      for (const diagnostic of diagnostics) {
        let severity = diagnostic.severity;
        const isUnnecessary = diagnostic.tags.includes('unnecessary');
        if (severity === 'hidden' && !isUnnecessary) continue;
        if (isUnnecessary && (severity === 'hidden' || severity === 'info')) severity = 'unnecessary'; // const range = this.#spanToRange(diagnostic.span);
        // annotations.push({
        //     severity,
        //     message: diagnostic.message,
        //     from: range.from,
        //     to: range.to,
        //     diagnostic
        // });
      } // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
      // this.#capturedUpdateLinting(this.#cm, annotations);
      // eslint-disable-next-line @typescript-eslint/no-non-null-assertion


      this[_Editor_options].on.slowUpdateResult({
        diagnostics: update.diagnostics,
        x: update.x
      });
    };

    this[_Editor_showConnectionLoss] = () => {
      if (!this[_Editor_connectionLossElement]) {
        const connectionLossElement = document.createElement('div');
        connectionLossElement.setAttribute('class', 'mirrorsharp-connection-issue');
        connectionLossElement.innerText = 'Server connection lost, reconnecting…';

        this[_Editor_wrapper].appendChild(connectionLossElement);

        this[_Editor_connectionLossElement] = connectionLossElement;
      }

      this[_Editor_wrapper].classList.add('mirrorsharp-connection-has-issue');
    };

    this[_Editor_hideConnectionLoss] = () => {
      this[_Editor_wrapper].classList.remove('mirrorsharp-connection-has-issue');
    };

    this[_Editor_hasDefaultServerOptions] = () => {
      const keys = Object.keys(this[_Editor_serverOptions]);
      return keys.length === 1 && keys[0] === 'language' && this[_Editor_serverOptions].language === defaultLanguage;
    };

    this[_Editor_sendServerOptions] = async value => {
      await this[_Editor_connection].sendSetOptions(value);
      await this[_Editor_requestSlowUpdate](true);
    };

    this[_Editor_receiveServerOptions] = value => {
      this[_Editor_serverOptions] = { ...this[_Editor_serverOptions],
        ...value
      }; // TODO: understand later
      // eslint-disable-next-line no-undefined

      if (value.language !== undefined && value.language !== this[_Editor_language]) {
        this[_Editor_language] = value.language; // this.#cm.setOption('mode', languageModes[this.#language]);
      }
    };

    this[_Editor_connection] = connection; // this.#selfDebug = selfDebug;

    options = {
      language: defaultLanguage,
      ...options,
      on: {
        slowUpdateWait: () => ({}),
        slowUpdateResult: () => ({}),
        textChange: () => ({}),
        connectionChange: () => ({}),
        serverError: message => {
          throw new Error(message);
        },
        ...options.on
      }
    };
    this[_Editor_options] = options; // const cmOptions = {
    //     gutters: [],
    //     indentUnit: 4,
    //     ...options.forCodeMirror,
    //     lineSeparator,
    //     // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
    //     mode: languageModes[options.language!],
    //     lint: { async: true, getAnnotations: this.#lintGetAnnotations, hasGutters: true },
    //     lintFix: { getFixes: this.#getLintFixes },
    //     infotip: { async: true, delay: 500, getInfo: this.#infotipGetInfo, render: renderInfotip }
    // } as CodeMirror.EditorConfiguration & { lineSeparator: string };
    // // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
    // cmOptions.gutters!.push('CodeMirror-lint-markers');
    // eslint-disable-next-line @typescript-eslint/no-non-null-assertion

    this[_Editor_language] = options.language;
    this[_Editor_serverOptions] = { ...(options.initialServerOptions ?? {}),
      language: this[_Editor_language]
    };

    if (!this[_Editor_hasDefaultServerOptions]()) {
      // eslint-disable-next-line @typescript-eslint/no-floating-promises
      this[_Editor_connection].sendSetOptions(this[_Editor_serverOptions]);
    }

    this[_Editor_shouldResendServerOptions] = false; // first open
    // const cmSource = (function getCodeMirror() {
    //     const next = textarea.nextSibling as { CodeMirror?: CodeMirror.EditorFromTextArea };
    //     if (next?.CodeMirror) {
    //         const existing = next.CodeMirror;
    //         for (const key in cmOptions) {
    //             // eslint-disable-next-line @typescript-eslint/no-explicit-any, @typescript-eslint/no-unsafe-call
    //             existing.setOption(key as any, cmOptions[key as keyof typeof cmOptions]);
    //         }
    //         return { cm: existing, existing: true };
    //     }
    //     return { cm: CodeMirror.fromTextArea(textarea, cmOptions) };
    // })();
    // this.#cm = cmSource.cm;
    // this.#keyMap = {
    //     /* eslint-disable object-shorthand */
    //     'Tab': () => {
    //         if (this.#cm.somethingSelected()) {
    //             this.#cm.execCommand('indentMore');
    //             return;
    //         }
    //         this.#cm.replaceSelection('    ', 'end');
    //     },
    //     'Shift-Tab': 'indentLess',
    //     // eslint-disable-next-line @typescript-eslint/no-floating-promises
    //     'Ctrl-Space': () => { connection.sendCompletionState('force'); },
    //     // eslint-disable-next-line @typescript-eslint/no-floating-promises
    //     'Shift-Ctrl-Space': () => { connection.sendSignatureHelpState('force'); },
    //     'Ctrl-.': 'lintFixShow',
    //     'Shift-Ctrl-Y': selfDebug ? () => {
    //         // eslint-disable-next-line @typescript-eslint/no-floating-promises
    //         selfDebug.requestData(connection);
    //     } : false
    //     /* eslint-enable object-shorthand */
    // };
    // this.#cm.addKeyMap(this.#keyMap);
    // // see https://github.com/codemirror/CodeMirror/blob/dbaf6a94f1ae50d387fa77893cf6b886988c2147/addon/lint/lint.js#L133
    // // ensures that next 'id' will be -1 whether a change happened or not
    // // eslint-disable-next-line @typescript-eslint/no-unsafe-member-access
    // this.#cm.state.lint.waitingFor = -2;
    // if (!cmSource.existing)
    //     this.setText(textarea.value);

    this[_Editor_wrapper] = document.createElement('div');

    this[_Editor_wrapper].classList.add('mirrorsharp');

    container.appendChild(this[_Editor_wrapper]);
    this[_Editor_cmView] = new EditorView({
      state: createState(this[_Editor_connection], {
        initialText: options.initialText,
        initialCursorOffset: options.initialCursorOffset
      })
    }); // if (selfDebug)
    //     selfDebug.watchEditor(this.getText, this.#getCursorIndex);

    this[_Editor_wrapper].appendChild(this[_Editor_cmView].dom); // const cmWrapper = this.#cm.getWrapperElement();
    // cmWrapper.classList.add('mirrorsharp', 'mirrorsharp-theme');


    this[_Editor_hinter] = new Hinter();
    this[_Editor_signatureTip] = new SignatureTip();
    this[_Editor_removeConnectionEvents] = addEvents(connection, {
      open: this[_Editor_onConnectionOpen],
      message: this[_Editor_onConnectionMessage],
      error: this[_Editor_onConnectionCloseOrError],
      close: this[_Editor_onConnectionCloseOrError]
    }); // this.#removeCodeMirrorEvents = addEvents(this.#cm, {
    //     beforeChange: this.#onCodeMirrorBeforeChange,
    //     cursorActivity: this.#onCodeMirrorCursorActivity,
    //     changes: this.#onCodeMirrorChanges
    // });
  }

  // #spanToRange = (span: SpanData) => {
  //     return {
  //         from: this.#cm.posFromIndex(span.start),
  //         to: this.#cm.posFromIndex(span.start + span.length)
  //     };
  // };
  // getCodeMirror() {
  //     return this.#cm;
  // }
  // setText(text: string) {
  //     this.#cm.setValue(text.replace(/(\r\n|\r|\n)/g, '\r\n'));
  // }
  getCodeMirrorView() {
    return this[_Editor_cmView];
  }

  getText() {
    return this[_Editor_cmView].state.doc.toString();
  }

  getLanguage() {
    return this[_Editor_language];
  }

  setLanguage(value) {
    return this[_Editor_sendServerOptions]({
      language: value
    });
  }

  setServerOptions(value) {
    return this[_Editor_sendServerOptions](value);
  }

  destroy(destroyOptions = {}) {
    // this.#cm.save();
    this[_Editor_removeConnectionEvents]();

    if (!destroyOptions.keepCodeMirror) {
      // this.#cm.toTextArea();
      return;
    } // this.#cm.removeKeyMap(this.#keyMap);
    // this.#removeCodeMirrorEvents();
    // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
    // this.#cm.setOption('lint', null!);
    // this.#cm.setOption('lintFix', null);
    // TODO: fix in infotip
    // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
    // this.#cm.setOption('infotip', null!);

  }

}