import type { EditorView } from '@codemirror/next/view';
import type { Language, DiagnosticSeverity } from './interfaces/protocol';
export declare type MirrorSharpLanguage = Language;
export declare type MirrorSharpConnectionState = 'open' | 'error' | 'close';
export interface MirrorSharpDiagnostic {
    readonly id: string;
    readonly severity: DiagnosticSeverity;
    readonly message: string;
}
export interface MirrorSharpSlowUpdateResult<TExtensionData = never> {
    readonly diagnostics: ReadonlyArray<MirrorSharpDiagnostic>;
    readonly x: TExtensionData;
}
export interface MirrorSharpOptions<TExtensionServerOptions = never, TSlowUpdateExtensionData = never> {
    readonly serviceUrl: string;
    readonly selfDebugEnabled?: boolean;
    readonly language?: MirrorSharpLanguage;
    readonly initialText?: string;
    readonly initialCursorOffset?: number;
    readonly on?: {
        readonly slowUpdateWait?: () => void;
        readonly slowUpdateResult?: (args: MirrorSharpSlowUpdateResult<TSlowUpdateExtensionData>) => void;
        readonly textChange?: (getText: () => string) => void;
        readonly connectionChange?: {
            (event: 'open', e: Event): void;
            (event: 'error', e: ErrorEvent): void;
            (event: 'close', e: CloseEvent): void;
        };
        readonly serverError?: (message: string) => void;
    };
    readonly initialServerOptions?: TExtensionServerOptions;
}
export interface MirrorSharpInstance<TExtensionServerOptions> {
    getCodeMirrorView(): EditorView;
    getText(): string;
    getLanguage(): MirrorSharpLanguage;
    setLanguage(value: MirrorSharpLanguage): void;
    setServerOptions(value: TExtensionServerOptions): Promise<void>;
    destroy(destroyOptions: {
        keepCodeMirror?: boolean;
    }): void;
}
export default function mirrorsharp<TExtensionServerOptions = never, TSlowUpdateExtensionData = never>(container: HTMLElement, options: MirrorSharpOptions<TExtensionServerOptions, TSlowUpdateExtensionData>): MirrorSharpInstance<TExtensionServerOptions>;
